/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Bitmap decompression routines
   Copyright (C) Matthew Chapman 1999-2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* jay.sorg@gmail.com */

/* indent is confused by this file */
/* *INDENT-OFF* */

#include "rdesktop.h"
#include "uimain.h"
#include "bitmap.h"

/* 24 to 32 bpp decompress */
#undef BM_FUNC_NAME
#undef BM_DEST_TYPE
#undef BM_IN_PIXEL_VAR
#undef BM_IN_PIXEL
#undef BM_BLACK
#undef BM_WHITE
#undef BM_ASSIGN_PIXEL
#undef BM_ASSIGN_PIXEL_ARRAY
#undef BM_ASSIGN_PIXEL_ARRAY_XOR
#undef BM_ASSIGN_BLACK
#undef BM_ASSIGN_WHITE
#undef BM_DEST_WIDTH_MULT
#undef BM_COLOUR1
#undef BM_COLOUR2
#undef BM_MIX
#undef BM_EXTRA
#define BM_FUNC_NAME bitmap_decompress_ex_24_to_32
#define BM_DEST_TYPE int
#define BM_IN_PIXEL_VAR(_p, _i) \
  b = CVAL(_i); g = CVAL(_i); r = CVAL(_i); \
  MAKE_COLOUR32(p, r, g, b); _p = p;
#define BM_IN_PIXEL(_p, _o, _i) \
  b = CVAL(_i); g = CVAL(_i); r = CVAL(_i); \
  MAKE_COLOUR32(p, r, g, b); _p[_o] = p;
#define BM_BLACK 0
#define BM_WHITE 0xffffff
#define BM_ASSIGN_PIXEL(_p1, _o1, _p2) _p1[_o1] = _p2;
#define BM_ASSIGN_PIXEL_ARRAY(_p1, _o1, _p2, _o2) _p1[_o1] = _p2[_o2];
#define BM_ASSIGN_PIXEL_ARRAY_XOR(_p1, _o1, _p2, _o2, _mix) \
  _p1[_o1] = (_mix) ^ (_p2[_o2]);
#define BM_ASSIGN_BLACK(_p, _o) _p[_o] = 0;
#define BM_ASSIGN_WHITE(_p, _o) _p[_o] = 0xffffff;
#define BM_DEST_WIDTH_MULT 1
#define BM_COLOUR1 colour1
#define BM_COLOUR2 colour2
#define BM_MIX mix
#define BM_EXTRA int p, r, g, b;
#include "bitmap_inc.c"

#undef BM_FUNC_NAME
#undef BM_DEST_TYPE
#undef BM_IN_PIXEL_VAR
#undef BM_IN_PIXEL
#undef BM_BLACK
#undef BM_WHITE
#undef BM_ASSIGN_PIXEL
#undef BM_ASSIGN_PIXEL_ARRAY
#undef BM_ASSIGN_PIXEL_ARRAY_XOR
#undef BM_ASSIGN_BLACK
#undef BM_ASSIGN_WHITE
#undef BM_DEST_WIDTH_MULT
#undef BM_COLOUR1
#undef BM_COLOUR2
#undef BM_MIX
#undef BM_EXTRA
#define BM_FUNC_NAME bitmap_decompress_ex_16_to_32
#define BM_DEST_TYPE int
#define BM_IN_PIXEL_VAR(_p, _i) \
  CVAL2(_i, p); SPLIT_COLOUR16(p, r, g, b); \
  MAKE_COLOUR32(p, r, g, b); _p = p;
#define BM_IN_PIXEL(_p, _o, _i) \
  CVAL2(_i, p); SPLIT_COLOUR16(p, r, g, b); \
  MAKE_COLOUR32(p, r, g, b); _p[_o] = p;
#define BM_BLACK 0
#define BM_WHITE 0xffffff
#define BM_ASSIGN_PIXEL(_p1, _o1, _p2) _p1[_o1] = _p2;
#define BM_ASSIGN_PIXEL_ARRAY(_p1, _o1, _p2, _o2) _p1[_o1] = _p2[_o2];
#define BM_ASSIGN_PIXEL_ARRAY_XOR(_p1, _o1, _p2, _o2, _mix) \
  _p1[_o1] = (_mix) ^ (_p2[_o2]);
#define BM_ASSIGN_BLACK(_p, _o) _p[_o] = 0;
#define BM_ASSIGN_WHITE(_p, _o) _p[_o] = 0xffffff;
#define BM_DEST_WIDTH_MULT 1
#define BM_COLOUR1 colour1
#define BM_COLOUR2 colour2
#define BM_MIX mix
#define BM_EXTRA int p, r, g, b;
#include "bitmap_inc.c"

RD_BOOL
bitmap_decompress_ex(uint8 * output, int width, int height, uint8 * input, int size,
                     int in_bpp, int out_bpp)
{
	if ((in_bpp == 16) && (out_bpp == 32))
	{
		return bitmap_decompress_ex_16_to_32(output, width, height, input, size);
	}
	else if ((in_bpp == 24) && (out_bpp == 32))
	{
		return bitmap_decompress_ex_24_to_32(output, width, height, input, size);
	}
	return False;
}

/* *INDENT-ON* */
