/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Bitmap decompression routines
   Copyright (C) Matthew Chapman 1999-2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This file should not be compiled directly.  It is included from
   another c file with certains defines in place. */
/* jay.sorg@gmail.com */

/* indent is confused by this file */
/* *INDENT-OFF* */

/* these things need to be defined
   BM_FUNC_NAME
   BM_DEST_TYPE
   BM_IN_PIXEL_VAR
   BM_IN_PIXEL
   BM_BLACK
   BM_WHITE
   BM_ASSIGN_PIXEL
   BM_ASSIGN_PIXEL_ARRAY
   BM_ASSIGN_PIXEL_ARRAY_XOR
   BM_ASSIGN_BLACK
   BM_ASSIGN_WHITE
   BM_DEST_WIDTH_MULT
   BM_COLOUR1
   BM_COLOUR2
   BM_MIX
   BM_EXTRA
   */

static RD_BOOL
BM_FUNC_NAME (uint8 *output, int width, int height, uint8 *input, int size)
{
	BM_DEST_TYPE *prevline = NULL, *line = NULL;
	BM_DEST_TYPE BM_COLOUR1 = BM_BLACK, BM_COLOUR2 = BM_BLACK;
	BM_DEST_TYPE BM_MIX = BM_WHITE;
	BM_EXTRA
	uint8 *end = input + size;
	int opcode, count, offset, isfillormix, x = width;
	int lastopcode = -1, insertmix = False, bicolour = False;
	uint8 code;
	uint8 mixmask, mask = 0;
	int fom_mask;

	while (input < end)
	{
		fom_mask = 0;
		code = CVAL(input);
		opcode = code >> 4;
		/* Handle different opcode forms */
		switch (opcode)
		{
			case 0xc:
			case 0xd:
			case 0xe:
				opcode -= 6;
				count = code & 0xf;
				offset = 16;
				break;
			case 0xf:
				opcode = code & 0xf;
				if (opcode < 9)
				{
					count = CVAL(input);
					count |= CVAL(input) << 8;
				}
				else
				{
					count = (opcode < 0xb) ? 8 : 1;
				}
				offset = 0;
				break;
			default:
				opcode >>= 1;
				count = code & 0x1f;
				offset = 32;
				break;
		}
		/* Handle strange cases for counts */
		if (offset != 0)
		{
			isfillormix = ((opcode == 2) || (opcode == 7));
			if (count == 0)
			{
				if (isfillormix)
				{
					count = CVAL(input) + 1;
				}
				else
				{
					count = CVAL(input) + offset;
				}
			}
			else if (isfillormix)
			{
				count <<= 3;
			}
		}
		/* Read preliminary data */
		switch (opcode)
		{
			case 0:	/* Fill */
				if ((lastopcode == opcode) &&
				    !((x == width) && (prevline == NULL)))
				{
					insertmix = True;
				}
				break;
			case 8:	/* Bicolour */
				BM_IN_PIXEL_VAR(colour1, input);
			case 3:	/* Colour */
				BM_IN_PIXEL_VAR(colour2, input);
				break;
			case 6:	/* SetMix/Mix */
			case 7:	/* SetMix/FillOrMix */
				BM_IN_PIXEL_VAR(mix, input);
				opcode -= 5;
				break;
			case 9:	/* FillOrMix_1 */
				mask = 0x03;
				opcode = 0x02;
				fom_mask = 3;
				break;
			case 0x0a:	/* FillOrMix_2 */
				mask = 0x05;
				opcode = 0x02;
				fom_mask = 5;
				break;
		}
		lastopcode = opcode;
		mixmask = 0;
		/* Output body */
		while (count > 0)
		{
			if (x >= width)
			{
				if (height <= 0)
				{
					return False;
				}
				x = 0;
				height--;
				prevline = line;
				line = ((BM_DEST_TYPE *) output) + height *
				       (width * BM_DEST_WIDTH_MULT);
			}
			switch (opcode)
			{
				case 0:	/* Fill */
					if (insertmix)
					{
						if (prevline == NULL)
						{
							BM_ASSIGN_PIXEL(line, x, mix);
						}
						else
						{
							BM_ASSIGN_PIXEL_ARRAY_XOR
							(line, x, prevline, x, mix);
						}
						insertmix = False;
						count--;
						x++;
					}
					if (prevline == NULL)
					{
						REPEAT(BM_ASSIGN_BLACK(line, x));
					}
					else
					{
						REPEAT(BM_ASSIGN_PIXEL_ARRAY
						(line, x, prevline, x))
					}
					break;
				case 1:	/* Mix */
					if (prevline == NULL)
					{
						REPEAT(BM_ASSIGN_PIXEL(line, x, mix));
					}
					else
					{
						REPEAT(BM_ASSIGN_PIXEL_ARRAY_XOR
						(line, x, prevline, x, mix));
					}
					break;
				case 2:	/* Fill or Mix */
					if (prevline == NULL)
					{
						REPEAT
						(
							MASK_UPDATE();
							if (mask & mixmask)
							{
								BM_ASSIGN_PIXEL(line, x, mix);
							}
							else
							{
								BM_ASSIGN_BLACK(line, x);
							}
						)
					}
					else
					{
						REPEAT
						(
							MASK_UPDATE();
							if (mask & mixmask)
							{
								BM_ASSIGN_PIXEL_ARRAY_XOR
								(line, x, prevline, x, mix);
							}
							else
							{
								BM_ASSIGN_PIXEL_ARRAY
								(line, x, prevline, x);
							}
						)
					}
					break;
				case 3:	/* Colour */
					REPEAT(BM_ASSIGN_PIXEL(line, x, colour2));
					break;
				case 4:	/* Copy */
					REPEAT(BM_IN_PIXEL(line, x, input));
					break;
				case 8:	/* Bicolour */
					REPEAT
					(
						if (bicolour)
						{
							BM_ASSIGN_PIXEL(line, x, colour2);
							bicolour = False;
						}
						else
						{
							BM_ASSIGN_PIXEL(line, x, colour1);
							bicolour = True;
							count++;
						}
					)
					break;
				case 0xd:	/* White */
					REPEAT(BM_ASSIGN_WHITE(line, x));
					break;
				case 0xe:	/* Black */
					REPEAT(BM_ASSIGN_BLACK(line, x));
					break;
				default:
					unimpl("bitmap opcode 0x%x\n", opcode);
					return False;
			}
		}
	}
	return True;
}

/* *INDENT-ON* */
